with Ada.Characters.Latin_1;
use  Ada.Characters.Latin_1;

function VSSL.Reals.Generic_String_to_Number (
	S    : String)
	return Number_Type
is
	function Is_Ignored (C : Character) return Boolean is (C in NUL .. ' ' | '_');

	function To_Upper (C : Character) return Character is
	(if C in 'a' .. 'z' then
		Character'Val (Character'Pos (C) - (Character'Pos ('a') - Character'Pos ('A')))
	else
		C);

	function Equal_No_Case (A, B : String) return Boolean is
	(A'Length = B'Length and then (for all J in A'Range =>
		To_Upper (A (J)) = To_Upper (B (J - A'First + B'First))));

	function Have_Affix (S, Prefix, Postfix : String) return Boolean is
	(S'Length > Prefix'Length + Postfix'Length and then
		Equal_No_Case (S (S'First .. S'First + Prefix'Length - 1), Prefix) and then
		Equal_No_Case (S (S'Last - Postfix'Length + 1 .. S'Last), Postfix));

	procedure Remove_Affix (F, L : in out Natural; Prefix, Postfix : in String) is
	begin
		F := F + Prefix'Length;
		L := L - Postfix'Length;
	end Remove_Affix;

	function Have_BOZ (S : String) return Boolean is
	(S'Length >= 4 and then S (S'First) in 'B' | 'b' | 'O' | 'o' | 'Z' | 'z' and then
	S (S'First + 1) = S (S'Last) and then S (S'Last) in ''' | '"'); --"

	-- Convert simple decimal number
	function Decimal (S : String) return Exponent_Type is
		N : Exponent_Type := 0;
		D : Exponent_Type;
	begin
		for C of S loop
			if C in '0' .. '9' then
				D := Character'Pos (C) - Character'Pos ('0');
				if N > (Exponent_Type'Last - D) / 10 then
					return Exponent_Type'First; -- carefully treat overflow
				end if; -- to have meaningfull exception
				N := N * 10 + D;
			elsif C not in NUL .. ' ' | '|' then
				raise Constraint_Error with "Illegal character " & Character'Image (C);
			end if;
		end loop;
		return N;
	end Decimal;

	-- Convert mantissa
	procedure Mantissa (
		S : in     String;
		R : in     Radix_Type;
		N :    out Number_Type;
		E :    out Exponent_Type)
	is
		Rdx : Number_Type := To_Number (R);
		Frac : Boolean := False;
		procedure Next_Digit (D : Integer) is
		begin
				if D >= R then
					raise Constraint_Error with "Illegal digit";
				end if;
				N := N * Rdx + To_Number (D);
				E := E - 1;
		end Next_Digit;
	begin
		N := To_Number (0);
		for J in S'Range loop
			if not Is_Ignored (S (J)) then
				case S (J) is
				when '0' .. '9' =>
					Next_Digit (Character'Pos (S (J)) - Character'Pos ('0'));
				when 'A' .. 'Z' =>
					Next_Digit (Character'Pos (S (J)) - Character'Pos ('A') + 10);
				when 'a' .. 'z' =>
					Next_Digit (Character'Pos (S (J)) - Character'Pos ('a') + 10);
				when '.' | ',' =>
					Frac := True;
					E := 0;
				when others =>
					raise Constraint_Error with "Invalid character " &
						Character'Image (S (J)) & " in number";
				end case;
			end if;
		end loop;
		if not Frac then
			E := 0;
		end if;
	end Mantissa;

	procedure Skip_Spaces (S : in String; J : in out Natural; L : in Natural) is
	begin
		while J <= L and then Is_Ignored (S (J)) loop
			J := J + 1;
		end loop;
		if J > L then
			raise Constraint_Error with "No number";
		end if;
	end Skip_Spaces;

	procedure Sign (S : in String; F : in out Natural; L : in Natural; Neg : out Boolean) is
	begin
		Neg := False;
		if S (F) = '+' then
			F := F + 1;
		elsif S (F) = '-' then
			F := F + 1;
			Neg := True;
		end if;
	end Sign;

	Radix  : Radix_Type := 10;
	MF     : Natural := S'First;
	ML     : Natural := S'Last;
	EF, EL,
	RF, RL : Natural := 0;
	Neg,                       -- Flag for negative
	Exp,                       -- Flag for exponential form
	Bin,                       -- Flag for binary exponent (when radix is not binary)
	Based,                     -- Flag for number with explicit radix
	NE     : Boolean := False; -- Flag for negative exponent
	EP,
	EE     : Exponent_Type := 0;
	N      : Number_Type := To_Number (0);

begin
	while ML >= MF and then Is_Ignored (S (ML)) loop -- skip
		ML := ML - 1;
	end loop;
	Skip_Spaces (S, MF, ML);

	Sign (S, MF, ML, Neg);
	Skip_Spaces (S, MF, ML);

	if Have_Affix (S (MF .. ML), "0X", "") then
		Radix := 16;  -- C-style 0x..
		Remove_Affix (MF, ML, "0X", "");
		Bin_Exp_Find : for J in MF + 1 .. ML - 1 loop
			Exp := S (J) in 'P' | 'p';
			if Exp then
				EL := ML;
				EF := J + 1;
				ML := J - 1;
				Bin := True;
				exit Bin_Exp_Find;
			end if;
		end loop Bin_Exp_Find;

	elsif Have_Affix (S (MF .. ML), "0B", "") then
		Radix := 2;    -- C-style 0b...
		Remove_Affix (MF, ML, "0B", "");
	elsif Have_Affix (S (MF .. ML), "B'", "'") or else Have_Affix (S (MF .. ML), "B""", """") then
		Radix := 2;
		Remove_Affix (MF, ML, "B'", "'");
	elsif Have_Affix (S (MF .. ML), "O'", "'") or else Have_Affix (S (MF .. ML), "O""", """") then
		Radix := 8;
		Remove_Affix (MF, ML, "O'", "'");
	elsif Have_Affix (S (MF .. ML), "Z'", "'") or else Have_Affix (S (MF .. ML), "Z""", """") then
		Radix := 16;
		Remove_Affix (MF, ML, "Z'", "'");
	else

		Radix_Find : for J in MF + 1 .. ML loop
			if S (J) in '#' | ':' then -- Ada-style based
				if not Based then -- first occurence of '#'
					RF := MF;
					RL := J - 1;
					MF := J + 1;
					Based := True;
				else -- second occurence of '#'
					if J < ML - 1 and then S (J + 1) in 'E' | 'e' then
						EL := ML;
						EF := J + 2;
						ML := J - 1;
						Exp := True;
						exit Radix_Find;
					elsif J = ML then
						ML := J - 1;
					end if;
				end if;
			elsif not Based and S (J) in 'R' | 'r' then -- Algol-style
				RF := MF;
				RL := J - 1;
				MF := J + 1;
				Based := True;
				exit Radix_Find;
			end if;
		end loop Radix_Find;

		if Based then
			Radix := Integer (Decimal (S (RF .. RL)));
			-- raise constraint error when not in range
		elsif not Based then
			Dec_Exp_Find : for J in MF + 1 .. ML - 1 loop
				Exp := S (J) in 'D' | 'd' | 'E' | 'e';
				if Exp then
					EL := ML;
					EF := J + 1;
					ML := J - 1;
					exit Dec_Exp_Find;
				end if;
			end loop Dec_Exp_Find;
		end if;
		if not Exp then
			if Have_Affix (S (MF .. ML), "", "B") then
				Radix := 2;
				Remove_Affix (MF, ML, "", "B");
			elsif Have_Affix (S (MF .. ML), "", "H") then
				Radix := 16;
				Remove_Affix (MF, ML, "", "B");
			end if;
		end if;
	end if;
	<<Common_Format>>

	Mantissa (S (MF .. ML), Radix, N, EP);

	if Exp then
		Sign (S, EF, EL, NE);
		EE := Decimal (S (EF .. EL));
		if EE < 0 then
			raise Constraint_Error with "Exponent value is out of range";
		end if;
		if NE then
			EE := -EE;
		end if;
	end if;

	if Bin then
		N := Multiply_by_Power (N, To_Number (Radix), EP);
		N := Multiply_by_Power (N, To_Number (2), EE);
	else
		N := Multiply_by_Power (N, To_Number (Radix), EP + EE);
	end if;

	if Neg then
		N := -N;
	end if;

	return N;

end VSSL.Reals.Generic_String_to_Number;